<?php

/* vim: set expandtab tabstop=4 shiftwidth=4 foldmethod=marker: */
// $HeadURL$
// +------------------------------------------------------------------------+
// | Maintain - DNS/DHCP Managment Tool                                     |
// +------------------------------------------------------------------------+
// | Copyright (c) 2005 Open Source University                              |
// +------------------------------------------------------------------------+
// | This program is free software; you can redistribute it and/or          |
// | modify it under the terms of the GNU General Public License            |
// | as published by the Free Software Foundation; either version 2         |
// | of the License, or (at your option) any later version.                 |
// |                                                                        |
// | This program is distributed in the hope that it will be useful,        |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of         |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          |
// | GNU General Public License for more details.                           |
// |                                                                        |
// | You should have received a copy of the GNU General Public License      |
// | along with this program; if not, write to the Free Software            |
// | Foundation, Inc., 59 Temple Place - Suite 330,                         |
// | Boston, MA  02111-1307, USA.                                           |
// +------------------------------------------------------------------------+
// | Author(s): Frederic Wenzel <fred@osuosl.org>                           |
// +------------------------------------------------------------------------+
// $Id$

/**
  * Contains the {@link Build_DNS} class
  *
  * @author $Author$
  * @filesource
  *
  * @package    classes
  * @subpackage backend
  */

require_once('PEAR.php');
require_once(MODDIR.'/module.php');
require_once(CLASSDIR.'/ip.php');

/**
 * Builds DNS config files and pushes them to the respective servers
 *
 * @package classes
 * @subpackage backend
 */
class Build_DNS extends Module
{
// {{{ vars
    var $log;                   // object history DB_DataObject
    var $conf;                  // module configuration
    var $buildfile;             // path to data file we are building to
// }}}

/* Methods for Config File Intergrity Checking */
// {{{ needsConfigFile()

    /**
      * Returns, if the module needs a config file at all. This will be used
      * by the automated config file integrity checks, for example.
      * Defaults to false.
      * Can be called statically.
      *
      * @return boolean .cfg file vital for the module to work?
      */
    function needsConfigFile()
    {
        return true;
    }
// }}}

/* Module Overloads */
// {{{ _install()

    function _install()
    {
        return true;
    }

// }}}
// {{{ _uninstall()

    function _uninstall()
    {
        return true;
    }

// }}}
// {{{ initialize()

    function initialize()
    {
        // what path are we building to?
        $this->buildfile = BUILD .'/'. $this->config['build_dns']['build_file'];
    
        return true;
    }

// }}}
// {{{ register_hooks()
/**
  * Connect this module to the module hooks it should reply to
  */
    function register_hooks()
    {
        Module::add_action('backend_build', 'buildDNS');
        Module::add_action('backend_push', 'pushDNS');
    }

// }}}


/* Private Module Functions */
// {{{ buildDNS()
/**
 * builds Maintain DNS configs
 * @return boolean success
 * @access private
 */

    function buildDNS()
    {
        // make a copy of the old config file that we can diff later
        if (file_exists($this->buildfile)) {
            if (!copy($this->buildfile, "{$this->buildfile}.bak"))
                PEAR::raiseError(_('Build_DNS could not duplicate dns config file. Check write permissions on build dir!'));
        }

        if (!$fh = @fopen($this->buildfile, 'w')) {     // try to open our config file for write access
            PEAR::raiseError(_("Build_DNS does not have write access to {$this->buildfile}!"));
            return false;
        }
        
        // now build DNS configs for the DNS server type specified
        $dnstype = strtolower($this->config['build_dns']['dns_type']);
        $dnstype_method = "_build_{$dnstype}";
        if (method_exists($this, $dnstype_method)) {
            $this->$dnstype_method($fh);
        } else {
            PEAR::raiseError(_("Build_DNS building of config files for DNS type {$dnstype} not supported!"));
        }

        fclose($fh);
    }
// }}}
// {{{ pushDNS()
/**
 * pushes DNS configs to the server(s)
 * @return boolean success
 * @access private
 */

    function pushDNS()
    {
        // only push to servers if configured to do so
        if (!$this->config['push_dns']['push']) return;
        
        // check conf file syntax if configured
        if ($this->config['push_dns']['check_syntax']) {
            if (!$this->check_current_config()) {
                $errormsg = (_("Push_DNS There are errors in the DNS data file. The file will not be pushed to the server. Please run bin/build/check-dns ASAP to find and correct the errors."));
                Email::adminEmail($errormsg, 'Push_DNS');
                PEAR::raiseError($errormsg);
                return false;
            }
        }

        // now push DNS configs for the DNS server type specified
        $dnstype = strtolower($this->config['build_dns']['dns_type']);
        $dnstype_method = "_push_{$dnstype}";
        if (method_exists($this, $dnstype_method)) {
            $this->$dnstype_method($fh);
        } else {
            PEAR::raiseError(_("Push_DNS pushing of config files for DNS type {$dnstype} not supported!"));
        }
    }
// }}}

// {{{ check_current_config()
/**
 * publicly accessible config to check current config file.
 * is used by /bin/build/check_dns, but also by push_dns
 *
 * @return boolean errors occured?
 * @access public
 */
    function check_current_config()
    {
        $dnstype = strtolower($this->config['build_dns']['dns_type']);
        $dnstype_method = "_{$dnstype}_check_conf_syntax";
        if (method_exists($this, $dnstype_method))
            return $this->$dnstype_method($this->buildfile);
        else
            return true;
    }
// }}}

// {{{ _build_tinydns()
/**
 * Build tinyDNS config files
 *
 * @param Filehandler pointing to dns config file to be built
 * @return void
 * @access private
 */
    function _build_tinydns(&$fh)
    {
        // include header template
        if (!$this->_include_file($fh, $this->_path . '/' . $this->config['tinydns']['header_file'])) {
            PEAR::raiseError(_("Build_DNS could not open tinydns header file for inclusion: ")." {$this->config['tinydns']['header_file']}!");
        }

        $this->_tiny_print_soa($fh);    // Start of Authorities
        $this->_tiny_print_mx($fh);     // Mail Exchangers
        $this->_tiny_print_cname($fh);  // CNames
        $this->_tiny_print_ptr($fh);    // Pointers
        $this->_tiny_print_ns($fh);     // Name Servers
        $this->_tiny_print_host($fh);   // Hosts
        $this->_tiny_print_dnscache();   // DNSCache configs

        // include footer template
        if (!$this->_include_file($fh, $this->_path . '/' . $this->config['tinydns']['footer_file'])) {
            PEAR::raiseError(_("Build_DNS could not open tinydns footer file for inclusion: ")." {$this->config['tinydns']['footer_file']}!");
        }
    }
// }}}
// {{{ _tiny_print_soa()
/**
 * fetches soa entries from the DB and prints them to the tinydns data file
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_soa(&$fh)
    {
        fwrite($fh, "# print_soa: These are the soa records for all of the\n");
        fwrite($fh, "#   domains we administer\n");
        
        $soa = DB_DataObject::Factory('soa');
        $domain = DB_DataObject::Factory('domain');
        $soa->joinAdd($domain);
        $soa->selectAdd('domain.name AS domainname');
        $soa->find();
        while ($soa->fetch()) {
            if ($soa->delegated == 'true') {
                fwrite($fh, "# Delegation of {$soa->domainname}\n");
            } else {
                fwrite($fh, "# soa & ns records for {$soa->domainname}\n");
                fwrite($fh, "Z{$soa->domainname}:{$soa->primary_master}:{$soa->hostmaster}::{$soa->refresh}:{$soa->retry}:{$soa->expire}:{$soa->ttl}:{$soa->ttl}::\n");
            }

            // now print the name server records for this domain
            $ns = DB_DataObject::Factory('nameserver');
            $ns->domain = $soa->domain;
            $ns->find();
            while ($ns->fetch()) {
                if ($soa->delegated == 'true') {
                    fwrite($fh, "\&{$soa->domainname}:{$ns->name}\n");
                } else {
                    fwrite($fh, ".{$soa->domainname}::{$ns->name}:86400::\n");
                }

                // if nameserver has a glue component, print it
                if ($ns->glue != 0) {
                    $glue_ip = IP::int2ip($ns->glue);
                    fwrite($fh, "+{$ns->name}::{$glue_ip}\n");
                }
                
            }
            $ns->free();

            fwrite($fh, "\n");
        }
        $domain->free();
        $soa->free();
        
        fwrite($fh, "\n\n");
    }
// }}}
// {{{ _tiny_print_mx()
/**
 * fetches MX entries from the DB and prints them to the tinydns data file
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_mx(&$fh)
    {
        global $conf;
        
        // select all the zones we have an mx for
        $mx_zones = DB_DataObject::Factory('mx');
        $zones = DB_DataObject::Factory('zone');
        $mx_zones->joinAdd($zones);
        $mx_zones->selectAdd();
        $mx_zones->selectAdd('DISTINCT mx.zone, zone.name');
        $mx_zones->orderBy('zone.name');
        $mx_zones->enabled = 1;
        $mx_zones->find();
        
        fwrite($fh, "# print_mx: Listing of all of the MX records for the domains\n");
        fwrite($fh, "#   we administrate\n");
        
        while ($mx_zones->fetch()) {
            fwrite($fh, "# MX for {$mx_zones->name}\n");

            // fetch MXes and their associated domain names
            $mx = DB_DataObject::Factory('mx');
            $domain = DB_DataObject::Factory('domain');
            $mx->joinAdd($domain);
            $mx->zone = $mx_zones->zone;
            $mx->enabled = 1;
            $mx->selectAdd();
            $mx->selectAdd('mx.*, domain.name AS domainname');
            $mx->find();

            while ($mx->fetch()) {
                $mxttl = ($mx->ttl == 0) ? $this->_check_ttl($conf['soa']['default_mx_ttl']) : $mx->ttl;

                fwrite ($fh, "@{$mx->name}.{$mx->domainname}::{$mx->server}:{$mx->priority}:{$mxttl}\n");
            }
            $domain->free();
            $mx->free();
            
            fwrite($fh, "\n");
            
        }
        $zones->free();
        $mx_zones->free();

        fwrite($fh, "\n\n");
    }
// }}}
// {{{ _tiny_print_cname()
/**
 * fetches CNAME entries from the DB and prints them to the tinydns data file
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_cname(&$fh)
    {
        global $conf;
        
        // select all the zones we have an mx for
        $cname_zones = DB_DataObject::Factory('cname');
        $zones = DB_DataObject::Factory('zone');
        $cname_zones->joinAdd($zones);
        $cname_zones->selectAdd();
        $cname_zones->selectAdd('DISTINCT cname.zone, zone.name');
        $cname_zones->orderBy('zone.name');
        $cname_zones->enabled = 1;
        $cname_zones->find();
        
        fwrite($fh, "# print_cname: Listing of all of the CNAME records for the domains\n");
        fwrite($fh, "#   we administrate\n");
        
        while ($cname_zones->fetch()) {
            fwrite($fh, "# CNAMEs for {$cname_zones->name}\n");

            // fetch CNAMEs and their associated domain names
            $cname = DB_DataObject::Factory('cname');
            $domain = DB_DataObject::Factory('domain');
            $cname->joinAdd($domain);
            $cname->zone = $cname_zones->zone;
            $cname->enabled = 1;
            $cname->selectAdd();
            $cname->selectAdd('cname.*, domain.name AS domainname');
            $cname->find();

            while ($cname->fetch()) {
                $cnamettl = ($cname->ttl == 0) ? $this->_check_ttl($conf['soa']['default_cname_ttl']) : $cname->ttl;

                fwrite ($fh, "C{$cname->name}.{$cname->domainname}:{$cname->server}:{$cnamettl}\n");
            }
            $domain->free();
            $cname->free();
            
            fwrite($fh, "\n");
            
        }
        $zones->free();
        $cname_zones->free();

        fwrite($fh, "\n\n");
    }
// }}}
// {{{ _tiny_print_ptr()
/**
 * fetches PTR entries from the DB and prints them to the tinydns data file
 * (either reverse or forward depending on 'type')
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_ptr(&$fh)
    {
        global $conf;
        
        // select all the zones we have pointers for
        $ptr = DB_DataObject::Factory('pointer');
        $ptr->orderBy('type');
        $ptr->enabled = 1;
        $ptr->find();
        
        fwrite($fh, "# print_ptr: These are all forward/reverse pointers we know of\n");
        fwrite($fh, "#\n");
        
        while ($ptr->fetch()) {
            $ptr_ip = IP::int2ip($ptr->ip);
            
            $ptrttl = ($ptr->ttl == 0) ? $this->_check_ttl($conf['soa']['default_host_ttl']) : $ptr->ttl;
            
            switch($ptr->type) {
            case 'forward':
                fwrite($fh, "+{$ptr->hostname}:{$ptr_ip}:{$ptrttl}\n");
                break;
                
            case 'reverse':
                // build reverse IP from forward IP
                $reverseip = explode('.', $ptr_ip);
                $reverseip = array_reverse($reverseip);
                $reverseip = implode('.', $reverseip);
                fwrite($fh, "^{$reverseip}.in-addr.arpa:{$ptr->hostname}:{$ptrttl}\n");
                break;
                
            default:
                PEAR::raiseError(_("Build_DNS unknown pointer type ")." {$ptr->type} "._(" for pointer id ")." {$ptr->id}");
                break;
            }
        }
        $ptr->free();

        fwrite($fh, "\n\n");
    }
// }}}
// {{{ _tiny_print_ns()
/**
 * Fetches name server entries from the DB and prints them to the tinydns data file.
 * We only print the NS records for things that don't have SOA records.
 * These are usually things we delegate off to other machines/hosts.
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_ns(&$fh)
    {
        // We only print the NS records for things that don't have SOA records.
        // These are usually things we delegate off to other machines/hosts.
        $ns = DB_DataObject::Factory('nameserver');
        $soa = DB_DataObject::Factory('soa');
        $domain = DB_DataObject::Factory('domain');
        
        $domain->joinAdd($soa, 'LEFT');
        $ns->joinAdd($domain);
        $ns->whereAdd('soa.id is NULL');
        $ns->selectAdd();
        $ns->selectAdd('nameserver.name AS nsname, domain.name AS domainname');
        $ns->find();
        
        fwrite($fh, "# print_ns: The following are NS records of domains we\n");
        fwrite($fh, "#   delegate to other servers\n");
        while ($ns->fetch()) {
            fwrite($fh, ".{$ns->domainname}::{$ns->nsname}:86400::\n");
        }
        $domain->free();
        $soa->free();
        $ns->free();
        fwrite($fh, "\n\n");
    }
// }}}
// {{{ _tiny_print_host()
/**
 * fetches host entries from the DB and prints them to the tinydns data file
 *
 * @param Filehandler pointing to the tinydns data file
 * @return void
 * @access private
 */
    function _tiny_print_host(&$fh)
    {
        global $conf;
        
        fwrite($fh, "# print_host: These are all of the hosts we have record of\n");
        fwrite($fh, "#\n");
        
        // fetch all hosts that have IP addresses assigned
        $hosts = DB_DataObject::Factory('host');
        $domain = DB_DataObject::Factory('domain');
        $hosts->joinAdd($domain);
        $hosts->whereAdd('ip != 0');
        $hosts->enabled = 1;
        $hosts->orderBy('ip');
        $hosts->selectAdd();
        $hosts->selectAdd('host.name, host.ip, host.ttl, domain.name AS domainname');
        $hosts->find();

        fwrite($fh, "# hosts with associated ip addresses\n");
        while ($hosts->fetch()) {
            $hostip = IP::int2ip($hosts->ip);
            $host_ttl = ($hosts->ttl == 0) ? $this->_check_ttl($conf['soa']['default_host_ttl']) : $hosts->ttl;
            
            fwrite($fh, "={$hosts->name}.{$hosts->domainname}:{$hostip}:{$host_ttl}\n");
        }
        $domain->free();
        $hosts->free();
        unset($hosts, $domain);

        // now go through all dynamic host ranges and print them out as well
        // quite complicated join: to get the dynamic range's domain name, we
        // join: ranges -> zone_range -> domain
        $dynranges = DB_DataObject::Factory('ranges');
        $zone_range = DB_DataObject::Factory('zone_range');
        $domain = DB_DataObject::Factory('domain');
        $zone_range->joinAdd($domain);
        $dynranges->joinAdd($zone_range);
        
        $dynranges->type = 'dynamic';
        $dynranges->enabled = 1;
        $dynranges->orderBy('start');
        $dynranges->selectAdd();
        $dynranges->selectAdd('ranges.start, ranges.end, domain.name AS domainname');
        $dynranges->find();

        fwrite($fh, "# dynamic ip address ranges\n");
        while ($dynranges->fetch()) {
            // walk through all ip addresses in this dynamic range
            // and make a host record for each of them
            for ($ip = $dynranges->start; $ip <= $dynranges->end; $ip++) {
                $hostip = IP::int2ip($ip);
                $hostname = str_replace('.', '-', $hostip);
                // default ttl for all dynamic host ranges
                // TODO check if this behaviour is appropriate
                $host_ttl = $this->_check_ttl($conf['soa']['default_host_ttl']);
                
                fwrite($fh, "={$hostname}.{$dynranges->domainname}:{$hostip}:{$host_ttl}\n");
            }
        }
        $domain->free();
        $zone_range->free();
        $dynranges->free();
        unset ($domain, $zone_range, $dynranges);
        
    }
// }}}
// {{{ _tiny_print_dnscache()
/**
 * Builds dnscache config files based on SOAs
 *
 * @return void
 * @access private
 */
    function _tiny_print_dnscache()
    {
        $templatedir = $this->_path.'/'.$this->config['dnscache']['dnscache_templates'];
        $outputdir = BUILD.'/dnscache';

        // see if the dnscache build dir is there or not
        // if so, empty it, if not, create it.
        if (is_dir($outputdir)) {
            foreach (glob($outputdir.'/*') as $outputfile) {
                unlink ($outputfile);
            }
        } else {
            mkdir ($outputdir);
        }

        // now fetch the domains for all of our SOAs and create a dnscache 
        // config pointing to localhost
        $soa = DB_DataObject::Factory('soa');
        $domain = DB_DataObject::Factory('domain');
        $soa->joinAdd($domain);
        $soa->selectAdd();
        $soa->selectAdd('DISTINCT domain.name');
        $soa->find();

        while ($soa->fetch()) {
            if (!$dnscachefile = @fopen($outputdir.'/'.$soa->name, 'w')) {
                PEAR::raiseError(_("Build_DNS could not open dnscache config file: ")." {$outputdir}/{$soa->name}!\n"._('dnscache configuration aborted!'));
                return false;
            }
            fwrite($dnscachefile, '127.0.0.1');
            fclose($dnscachefile);
        }
        
        $domain->free();
        $soa->free();
    }
// }}}
// {{{ _tinydns_check_conf_syntax()
/**
 * checks the syntax of the tinydns configuration file, using the valtz
 * script (http://x42.com/software/valtz/) bundled to this module
 *
 * @param string buildfile path
 * @return boolean success
 * @access private
 */
    function _tinydns_check_conf_syntax($checkfile)
    {
        global $mode;   // tells us if debug mode is on
        
        // where is valtz?
        $valtz = "{$this->_path}/valtz/valtz";
        
        if ($mode['debug'] === true) {
            system("$valtz -x $checkfile", $checkresult);
        } else {
            if ($this->config['push_dns']['ignore_char_errors'] == true) {
                // check for everything but character errors in names
                $checkresult = exec("$valtz $checkfile 2>&1 | grep pos | grep -vc ASCII");
            } else {
                system("$valtz -x $checkfile > /dev/null 2>&1", $checkresult);
            }
        }
        
        // return if errors have occured or not
        return ($checkresult == 0);
    }
// }}}
// {{{ _push_tinydns()
/**
 * push configs to tinydns servers and dnscache servers as configured
 *
 * @return boolean success
 * @access private
 */
    function _push_tinydns()
    {
        global $mode;
    
        // fetch all configured servers
        $config_keys = array_keys($this->config['tinydns']);
        $tiny_servers = preg_grep("/^tinydns_server_.*$/", $config_keys);
        $dnsc_servers = preg_grep("/^dnscache_server_.*$/", $config_keys);
        
        // now cycle through all tinydns servers and push the configs out!
        foreach ($tiny_servers as $server_id) {
            // extract unique identifier out of server_id string
            $server_id = preg_replace("/^(tinydns_server_)/", '', $server_id);

            $location = $this->config['tinydns']["tinydns_server_{$server_id}"];
            $rsync_extra = (isset($this->config['tinydns']["tinydns_rsync_extra_{$server_id}"]) ? 
                    $this->config['tinydns']["tinydns_rsync_extra_{$server_id}"] : '');

            // build our rsync command
            $rsync = "rsync -a ";
            $rsync .= $rsync_extra;
            
            exec("$rsync {$this->buildfile} $location 2>&1", $output, $result);

            if ($result != 0) { // error
                $errormsg = (_("Push_DNS Error while pushing tinydns configs to server ") . $location);
                $this->_push_error($errormsg, $output);
            }
        }

        // now do the same for the dnscache servers
        foreach ($dnsc_servers as $server_id) {
            // extract unique identifier out of server_id string
            $server_id = preg_replace("/^(dnscache_server_)/", '', $server_id);

            $location = $this->config['tinydns']["dnscache_server_{$server_id}"];
            $rsync_extra = (isset($this->config['tinydns']["dnscache_rsync_extra_{$server_id}"]) ? 
                    $this->config['tinydns']["dnscache_rsync_extra_{$server_id}"] : '');

            // build our rsync command
            $rsync = "rsync -a ";
            $rsync .= $rsync_extra;
            
            exec("$rsync --delete --exclude \"@\" " .BUILD. "/dnscache/ {$this->config['dnscache']['dnscache_templates']} $location 2>&1", $output, $result);

            if ($result != 0) { // error
                $errormsg = (_("Push_DNS Error while pushing dnscache configs to server ") . $location);
                $this->_push_error($errormsg, $output);
            }
        }

        // execute post-push command (for restarting services and such)
        if ($this->config['tinydns']['post_push']) {
            exec($this->config['tinydns']['post_push'].' 2>&1', $output, $result);
            if ($result != 0) {
                $errormsg = _("Push_DNS Error while executing post-push scripts.");
                $this->_push_error($errormsg, $output);
            }
        }
    }
// }}}

// {{{ _include_file
/**
 * include a template file into the output data file
 *
 * @param filehandler of output file
 * @param string filename of file to be included
 * @return boolean success
 * @access private
 */
    function _include_file(&$fh, $incfile)
    {
        // try to open include template and put it into the config file
        if (!$inc_fh = @fopen($incfile, 'r')) {
            return false;
        } else {
            while (!feof($inc_fh)) {
                fwrite($fh, fgets($inc_fh));
            }
            fclose($inc_fh);
            return true;
        }
        
    }
// }}}
// {{{ _push_error()
/**
  * If there was a push error, notify admins and throw a PEAR error.
  *
  * @access private
  * @param Error message
  * @param Output of the command that failed
  * @return void
  */
    function _push_error($msg, $output = '')
    {
        if ($output) {
            $msg .= "\n\n";
            foreach ($output as $line) $msg .= $line . "\n";
        }
        Email::adminEmail($msg, "Push_DNS");
        PEAR::raiseError($msg);
    }
// }}}

// {{{ _check_ttl()
/** 
 * checks for a given ttl value to be larger than 300
 * it's necessary because "BIND currently considers any
 * TTL under 300 seconds as 'irrational'",
 * see: http://www.ietf.org/rfc/rfc1794.txt
 * TODO: should this restriction be used for tinydns also?
 *
 * @param int TTL value to be checked
 * @return int 300 if parameter was less than 300, unchanged TTL else
 * @access private
 */
    function _check_ttl($ttl) {
        return ($ttl < 300) ? 300 : $ttl;
    }
// }}}

}

?>
